@charset utf-8;

/* Shared CSS for AsciiDoc xhtml11 and html5 backends */

/* Dave Mills's CSS definitions first */
body {
  background: #fdf1e1;
  color: #060;
  font-family: "verdana", sans-serif;
  margin-left: 5px;
  text-align: justify;
}

p,
h4,
hr,
li {
  margin-top: 0.6em;
  margin-bottom: 0.6em;
}

li.inline {
  text-align: left;
  margin-top: 0;
  margin-bottom: 0;
}

ul,
dl,
ol {
  margin-top: 0.6em;
  margin-bottom: 0.6em;
  margin-left: 5em;
}

dt {
  margin-top: 0.6em;
  color: #060;
  font-style: normal;
  margin-bottom: 0;
  /* margin-top: 0.5em; */
  /* color: navy; */
}

dd {
  margin-bottom: 0.6em;
}

div.header {
  font-style: italic;
  text-align: center;
}

div.footer {
  font-size: 60%;
  text-align: center;
}

img.cell {
  /* align: left; */
}

td.sidebar {
  width: 40px;
/*
  align: center;
  valign: top;
*/
}

img.sidebar {
  /* align: center; */
  margin-top: 5px;
}

h4.sidebar {
  /* align: center; */
}

p.top {
  background: #fdf1e1;
  color: #060;
  margin-left: -90px;
  position: absolute;
  text-align: center;
}

a:link.sidebar {
  background: transparent;
  color: #903;
  font-weight: bold;
}

a:visited.sidebar {
  background: transparent;
  color: #903;
  font-weight: bold;
}

a:hover.sidebar {
  background: #fdf1e1;
  color: #060;
}

img {
  margin: 5px;
}

/* div {
  text-align: center;
}

*/
h1 {
  font-size: 250%;
  text-align: center;
}

caption {
  background: #eee;
  color: #399;
}

/*
tx {
  text-align: center;
}
*/

th {
  background: #ffc;
  color: #060;
  padding-top: 5px;
  text-align: center;
  text-decoration: underline;
}

th.caption {
  background: #eee;
  color: #060;
  text-align: center;
}

/* hacked version of asciidoc stylesheet */

/* remove asciidoc's body declaration */

/* Title font. */
h1,
h2,
h3,
h4,
h5,
h6,
div.title,
caption.title,
thead,
p.table.header,
#toctitle,
#author,
#revnumber,
#revdate,
#revremark,
#footer {
  font-family: Arial, Helvetica, sans-serif;
}

/* remove asciidoc's body margin declaration */

/* remove asciidoc's messing with the link colors */

/* remove asciidoc's styling for em */

/* remove asciidoc's styling for strong */

/* use asciidoc's header styling with Dave's color */
h1,
h2,
h3,
h4,
h5,
h6 {
  color: #060;
  line-height: 1.3;
  margin-bottom: 0.5em;
  margin-top: 1.2em;
}

h1,
h2,
h3 {
  border-bottom: 2px solid silver;
}

h2 {
  padding-top: 0.5em;
}

h3 {
  float: left;
}

h3 + * {
  clear: left;
}

h5 {
  font-size: 1em;
}

div.sectionbody {
  margin-left: 0;
}

hr {
  border: 1px solid silver;
}

p {
  margin-bottom: 0.5em;
  margin-top: 0.5em;
}

ul,
ol,
li > p {
  margin-top: 0;
}

/*
ul > li {
  color: #aaa;
}

ul > li > * {
  color: black;
}
*/

.monospaced,
code,
pre {
  color: #006600;
  font-family: "Courier New", Courier, monospace;
  font-size: inherit;
  margin: 0;
  padding: 0;

  /* color: navy; */
}

pre {
  white-space: pre-wrap;
}

#author {
  font-size: 1.1em;
  font-weight: bold;

  /* color: #527bbd; */
}

#footer {
  border-top: 2px solid silver;
  font-size: small;
  margin-top: 4em;
  padding-top: 0.5em;
}

#footer-text {
  float: left;
  padding-bottom: 0.5em;
}

#footer-badges {
  float: right;
  padding-bottom: 0.5em;
}

#preamble {
  margin-bottom: 1.5em;
  margin-top: 1.5em;
}

div.imageblock,
div.exampleblock,
div.verseblock,
div.quoteblock,
div.literalblock,
div.listingblock,
div.sidebarblock,
div.admonitionblock {
  margin-bottom: 1.5em;
  margin-top: 1em;
}

div.admonitionblock {
  margin-bottom: 2em;
  margin-right: 10%;
  margin-top: 2em;

  /* color: #606060; */
}

div.content {
  /* Block element content. */
  padding: 0;
}

/* Block element titles. */
div.title,
caption.title {
  font-weight: bold;
  margin-bottom: 0.5em;
  margin-top: 1em;
  text-align: left;

  /* color: #527bbd; */
}

div.title + * {
  margin-top: 0;
}

td div.title:first-child {
  margin-top: 0;
}

div.content + div.title {
  margin-top: 0;
}

div.content div.title:first-child {
  margin-top: 0;
}

div.sidebarblock > div.content {
  background: #ffe;
  border: 1px solid #ddd;
  border-left: 4px solid #f0f0f0;
  padding: 0.5em;
}

div.listingblock > div.content {
  background: #f8f8f8;
  border: 1px solid #ddd;
  border-left: 5px solid #f0f0f0;
  padding: 0.5em;
}

div.quoteblock,
div.verseblock {
  border-left: 5px solid #f0f0f0;
  color: #888;
  margin-left: 1em;
  margin-right: 10%;
  padding-left: 1em;
}

div.quoteblock > div.attribution {
  padding-top: 0.5em;
  text-align: right;
}

div.verseblock > pre.content {
  font-family: inherit;
  font-size: inherit;
}

div.verseblock > div.attribution {
  padding-top: 0.75em;
  text-align: left;
}

/* DEPRECATED: Pre version 8.2.7 verse style literal block. */
div.verseblock + div.attribution {
  text-align: left;
}

div.admonitionblock .icon {
  color: #527bbd;
  font-size: 1.1em;
  font-weight: bold;
  padding-right: 0.5em;
  text-decoration: underline;
  vertical-align: top;
}

div.admonitionblock td.content {
  border-left: 3px solid #ddd;
  padding-left: 0.5em;
}

div.exampleblock > div.content {
  border-left: 3px solid #ddd;
  padding-left: 0.5em;
}

div.imageblock div.content {
  padding-left: 0;
}

span.image img {
  border-style: none;
  vertical-align: text-bottom;
}

a.image:visited {
  color: white;
}

dl {
  margin-bottom: 0.8em;
  margin-top: 0.8em;
}

dd > *:first-child {
  margin-top: 0.1em;
}

ul,
ol {
  list-style-position: outside;
}

ol.arabic {
  list-style-type: decimal;
}

ol.loweralpha {
  list-style-type: lower-alpha;
}

ol.upperalpha {
  list-style-type: upper-alpha;
}

ol.lowerroman {
  list-style-type: lower-roman;
}

ol.upperroman {
  list-style-type: upper-roman;
}

div.compact ul,
div.compact ol,
div.compact p,
div.compact div {
  margin-bottom: 0.1em;
  margin-top: 0.1em;
}

tfoot {
  font-weight: bold;
}

td > div.verse {
  white-space: pre;
}

div.hdlist {
  margin-bottom: 0.8em;
  margin-top: 0.8em;
}

div.hdlist tr {
  padding-bottom: 15px;
}

dt.hdlist1.strong,
td.hdlist1.strong {
  font-weight: bold;
}

td.hdlist1 {
  font-style: normal;
  padding-right: 0.8em;
  vertical-align: top;

  /* color: navy; */
}

td.hdlist2 {
  vertical-align: top;
}

div.hdlist.compact tr {
  margin: 0;
  padding-bottom: 0;
}

.comment {
  background: yellow;
}

.footnote,
.footnoteref {
  font-size: 0.8em;
}

span.footnote,
span.footnoteref {
  vertical-align: super;
}

#footnotes {
  margin: 20px 0 20px 0;
  padding: 7px 0 0 0;
}

#footnotes div.footnote {
  margin: 0 0 5px 0;
}

#footnotes hr {
  border: none;
  border-top: 1px solid silver;
  height: 1px;
  margin-left: 0;
  min-width: 100px;
  text-align: left;
  width: 20%;
}

div.colist td {
  padding-bottom: 0.3em;
  padding-right: 0.5em;
  vertical-align: top;
}

div.colist td img {
  margin-top: 0.3em;
}

@media print {
  #footer-badges {
    display: none;
  }
}

#toc {
  margin-bottom: 2.5em;
}

#toctitle {
  font-size: 1.1em;
  font-weight: bold;
  margin-bottom: 0.1em;
  margin-top: 1em;

  /* color: #527bbd; */
}

div.toclevel0,
div.toclevel1,
div.toclevel2,
div.toclevel3,
div.toclevel4 {
  margin-bottom: 0;
  margin-top: 0;
}

div.toclevel2 {
  font-size: 0.9em;
  margin-left: 2em;
}

div.toclevel3 {
  font-size: 0.9em;
  margin-left: 4em;
}

div.toclevel4 {
  font-size: 0.9em;
  margin-left: 6em;
}

span.aqua {
  color: aqua;
}

span.black {
  color: black;
}

span.blue {
  color: blue;
}

span.fuchsia {
  color: fuchsia;
}

span.gray {
  color: gray;
}

span.green {
  color: green;
}

span.lime {
  color: lime;
}

span.maroon {
  color: maroon;
}

span.navy {
  color: navy;
}

span.olive {
  color: olive;
}

span.purple {
  color: purple;
}

span.red {
  color: red;
}

span.silver {
  color: silver;
}

span.teal {
  color: teal;
}

span.white {
  color: white;
}

span.yellow {
  color: yellow;
}

span.aqua-background {
  background: aqua;
}

span.black-background {
  background: black;
}

span.blue-background {
  background: blue;
}

span.fuchsia-background {
  background: fuchsia;
}

span.gray-background {
  background: gray;
}

span.green-background {
  background: green;
}

span.lime-background {
  background: lime;
}

span.maroon-background {
  background: maroon;
}

span.navy-background {
  background: navy;
}

span.olive-background {
  background: olive;
}

span.purple-background {
  background: purple;
}

span.red-background {
  background: red;
}

span.silver-background {
  background: silver;
}

span.teal-background {
  background: teal;
}

span.white-background {
  background: white;
}

span.yellow-background {
  background: yellow;
}

span.big {
  font-size: 2em;
}

span.small {
  font-size: 0.6em;
}

span.underline {
  text-decoration: underline;
}

span.overline {
  text-decoration: overline;
}

span.line-through {
  text-decoration: line-through;
}

div.unbreakable {
  page-break-inside: avoid;
}

/* * xhtml11 specific * * */
div.tableblock {
  margin-bottom: 1.5em;
  margin-top: 1em;
}

div.tableblock > table {
  border: 3px solid #527bbd;
}

thead,
p.table.header {
  font-weight: bold;

  /* color: #527bbd;
  */
}

p.table {
  margin-top: 0;
}

/* Because the table frame attribute is overridden by CSS in most browsers. */
div.tableblock > table[frame="void"] {
  border-style: none;
}

div.tableblock > table[frame="hsides"] {
  border-left-style: none;
  border-right-style: none;
}

div.tableblock > table[frame="vsides"] {
  border-bottom-style: none;
  border-top-style: none;
}

/* * html5 specific * * */
table.tableblock {
  margin-bottom: 1.5em;
  margin-top: 1em;
  border-collapse: collapse;
  border-color: #527bbd;
  border-spacing: 0;
  border-style: solid;
  border-width: 3px;
}

thead,
p.tableblock.header {
  color: #527bbd;
  font-weight: bold;
}

p.tableblock {
  margin-top: 0;
}

th.tableblock,
td.tableblock {
  border-color: #527bbd;
  border-style: solid;
  border-width: 1px;
  padding: 4px;
}

table.tableblock.frame-topbot {
  border-left-style: hidden;
  border-right-style: hidden;
}

table.tableblock.frame-sides {
  border-bottom-style: hidden;
  border-top-style: hidden;
}

table.tableblock.frame-none,
table.tableblock.frame-none td {
  border-style: hidden;
}

th.tableblock.halign-left,
td.tableblock.halign-left {
  text-align: left;
}

th.tableblock.halign-center,
td.tableblock.halign-center {
  text-align: center;
}

th.tableblock.halign-right,
td.tableblock.halign-right {
  text-align: right;
}

th.tableblock.valign-top,
td.tableblock.valign-top {
  vertical-align: top;
}

th.tableblock.valign-middle,
td.tableblock.valign-middle {
  vertical-align: middle;
}

th.tableblock.valign-bottom,
td.tableblock.valign-bottom {
  vertical-align: bottom;
}

/* * manpage specific * * */
body.manpage h1 {
  border-bottom: 2px solid silver;
  border-top: 2px solid silver;
  padding-bottom: 0.5em;
  padding-top: 0.5em;
}

body.manpage h2 {
  border-style: none;
}

body.manpage div.sectionbody {
  margin-left: 3em;
}

@media print {
  body.manpage div#toc {
    display: none;
  }
}
