/*
 * Copyright Frank Kardel <kardel <AT> ntp.org>
 * Copyright the NTPsec project contributors
 * SPDX-License-Identifier: BSD-3-Clause
 */
#ifndef GUARD_ASCII_H
#define GUARD_ASCII_H

/*
 * just name the common ASCII control codes
 */
#define NUL	  0
#define STX	  2
#define ETX	  3
#define SOH	  1
#define DLE	 16

#endif
