#ifndef GUARD_TESTS_MAIN_H
#define GUARD_TESTS_MAIN_H

#include "config.h"

#include "unity.h"
#include "unity_fixture.h"

#include "ntp_stdlib.h"
#include "ntp_auth.h"

const char* tests_main_args(int arg);

#endif // GUARD_TESTS_MAIN_H
