# -*- coding: utf-8 -*-
#
# This code runs compatibly under Python 2 and 3.x for x >= 2.
# Preserve this property!

# Copyright the NTPsec project contributors
#
# SPDX-License-Identifier: BSD-2-Clause

from __future__ import absolute_import  # Ensure Python2 behaves like Python 3

api_major_version = 1   # bumped on incompatible changes
api_minor_version = 0   # bumped on compatible changes
