#define RELEASE_DATE 1703823396

/* We would like to use the build date.
 * But that breaks repeatable builds.

 * Update this before each release.
 * Some tests may need to be updated.
 * So when the release is announced is a good time.

 * Get the magic number from "date +%s"

 */
